/*
 * Copyright (c) 2022-present Charles7c Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.zayac.admin.service;

import cn.hutool.core.lang.TypeReference;
import cn.hutool.json.JSONUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.zayac.admin.resp.ApiResponse;
import com.zayac.admin.system.model.resp.AccountResp;
import io.netty.handler.timeout.TimeoutException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpStatusCode;
import org.springframework.stereotype.Service;
import org.springframework.web.reactive.function.BodyInserters;
import org.springframework.web.reactive.function.client.WebClient;
import org.springframework.web.reactive.function.client.WebClientResponseException;
import reactor.core.publisher.Mono;
import reactor.core.scheduler.Schedulers;
import reactor.util.retry.Retry;
import top.continew.starter.core.exception.BusinessException;

import java.time.Duration;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

@Service
@Slf4j
public class CompletableFutureWebClientService {
    private final WebClient webClient;
    private final Semaphore semaphore;
    private final ObjectMapper objectMapper;

    public CompletableFutureWebClientService(WebClient.Builder webClientBuilder,
                                             @Value("${webclient.max-concurrent-requests}") int maxConcurrentRequests,
                                             ObjectMapper objectMapper) {
        this.webClient = webClientBuilder.build();
        this.semaphore = new Semaphore(maxConcurrentRequests);
        this.objectMapper = objectMapper;
    }

    public <T> CompletableFuture<T> fetchDataForAccount(AccountResp account,
                                                        String apiPath,
                                                        Object params,
                                                        ParameterizedTypeReference<ApiResponse<T>> typeRef) {
        // 创建一个CompletableFuture列表，每个元素对应一个API请求
        return this.fetchData(account.getPlatformUrl() + apiPath, account.getHeaders(), params, typeRef);
    }

    //    public <T> CompletableFuture<T> fetchData(String url, String headers, Object params, ParameterizedTypeReference<ApiResponse<T>> typeRef) {
    //        webClient.mutate()
    //                .filters(filters -> filters.add(0, new ApiResponseHandler<>(typeRef))) // 添加到过滤器链的开始
    //                .build();
    //        System.out.println(webClient);
    //        return webClient.post()
    //                .uri(url)
    //
    //                .headers(httpHeaders -> {
    //                    Map<String, String> headerMap = JSONUtil.toBean(headers, new TypeReference<>() {
    //                    }, true);
    //                    headerMap.forEach(httpHeaders::add);
    //                })
    //                .body(params != null ? BodyInserters.fromValue(params) : BodyInserters.empty())
    //                .retrieve()
    //                .bodyToMono(typeRef)
    //                .map(ApiResponse::data)
    //                .doFinally(signal -> semaphore.release())
    //                .toFuture();
    //    }

    public <T> CompletableFuture<T> fetchData(String url,
                                              String headers,
                                              Object params,
                                              ParameterizedTypeReference<ApiResponse<T>> typeRef) {
        return Mono.fromCallable(() -> {
            if (!semaphore.tryAcquire(10, TimeUnit.SECONDS)) {  // 使用非阻塞方式获取许可
                throw new RuntimeException("Unable to acquire a permit");
            }
            return true;
        }).subscribeOn(Schedulers.boundedElastic()).then(this.webClient.post().uri(url).headers(httpHeaders -> {
            Map<String, String> headerMap = JSONUtil.toBean(headers, new TypeReference<>() {
            }, true);
            headerMap.forEach(httpHeaders::add);
        })
            .body(params != null ? BodyInserters.fromValue(params) : BodyInserters.empty())
            .retrieve()
            .onStatus(HttpStatusCode::isError, response -> Mono.error(new BusinessException("API call failed")))
            .bodyToMono(String.class)
            .doOnNext(resStr -> {
                log.info("request url:{}", url);
                log.info("request headers :{}", headers);
                log.info("request params:{}", params);
                log.info("response {}", resStr);
            })
            .flatMap(body -> {
                try {
                    ApiResponse<T> apiResponse = objectMapper.readValue(body, objectMapper.getTypeFactory()
                        .constructType(typeRef.getType()));
                    return Mono.justOrEmpty(apiResponse);
                } catch (Exception e) {
                    log.warn("JSON parsing exception: " + e.getMessage());
                    return Mono.just(new ApiResponse<T>(null, "Decoding error", 6008));
                }
            })
            .flatMap(this::respHandler)
            .retryWhen(Retry.backoff(3, Duration.ofSeconds(3)).filter(this::isRetryableException)))
            .doFinally(signal -> semaphore.release())
            .toFuture();
    }

    private boolean isRetryableException(Throwable throwable) {
        // 判断错误类型，只对特定的错误进行重试
        return throwable instanceof TimeoutException || throwable instanceof WebClientResponseException.ServiceUnavailable;
    }

    private <T> Mono<T> respHandler(ApiResponse<T> response) {
        if (response.getStatusCode().equals(6000)) {
            return Mono.just(response.getData());
        } else if (response.getStatusCode().equals(6001)) {
            // 重新登录逻辑或其他处理
            return Mono.error(new BusinessException("API token expired"));
        } else {
            return Mono.error(new BusinessException("Error status code: " + response.getStatusCode()));
        }
    }
}